# -*- coding: utf-8 -*-
"""Physical Constants.

A set of physical constants used across PyBlobby3D.

@author: Mathew Varidel

"""


class PhysicalConstants:
    """Physical constants required for Blobby3D."""
    C = 2.99792458e5  # speed of light (km/s)

    # Vacuum emission line wavelengths.
    HA = 6562.81  # HAlpha (Angstrom)
    N2Low = 6548.1  # Lower NII line
    N2Upp = 6583.1  # Upper NII line
